;; Generated package description from nix-mode.el  -*- no-byte-compile: t -*-
(define-package "nix-mode" "1.5.0.0.20230421.153655" "Major mode for editing .nix files" '((emacs "25.1") (magit-section "0") (transient "0.3")) :commit "719feb7868fb567ecfe5578f6119892c771ac5e5" :maintainer '("Matthew Bauer" . "mjbauer95@gmail.com") :keywords '("nix" "languages" "tools" "unix") :url "https://github.com/NixOS/nix-mode")
